/*-----------------------------------------------------------------------------------------------------*
 * UNIVERSIDADE DO VALE DO RIO DOS SINOS - MESTRADO EM COMPUTACAO APLICADA                             *
 * Sistemas Adaptativos Inteligentes - Prof. Fernando Santos Osorio                                    *
 * Autor: Milton Roberto Heinen - Matr. 392201                                                         *
 *-----------------------------------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------------------------------*
 * Includes                                                                                            *
 *-----------------------------------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include "recipes.h"


/*-----------------------------------------------------------------------------------------------------*
 * Defines                                                                                             *
 *-----------------------------------------------------------------------------------------------------*/
#define NRANSI
#define NR_END 1
#define FREE_ARG char*
#define ITMAX_POWELL 20
#define ITMAX_BRENT 10
#define TOL 2.0e-4
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define CGOLD 0.3819660
#define ZEPS 1.0e-10


/*-----------------------------------------------------------------------------------------------------*
 * Macros                                                                                              *
 *-----------------------------------------------------------------------------------------------------*/
static double dsqrarg;
#define DSQR(a) ((dsqrarg=(a)) == 0.0 ? 0.0 : dsqrarg*dsqrarg)

static double dmaxarg1, dmaxarg2;
#define DMAX(a,b) (dmaxarg1=(a),dmaxarg2=(b),(dmaxarg1) > (dmaxarg2) ? (dmaxarg1) : (dmaxarg2))

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);


/*-----------------------------------------------------------------------------------------------------*
 * Variaveis globais                                                                                   *
 *-----------------------------------------------------------------------------------------------------*/
int ncom, iTamanhoMaximo = 0;
double *pcom = NULL, *xicom = NULL, (*nrfunc) (double[]);
double *pt = NULL, *ptt = NULL, *xit = NULL, *xt = NULL;


/*-----------------------------------------------------------------------------------------------------*
 * Prototipos locais                                                                                   *
 *-----------------------------------------------------------------------------------------------------*/
inline void nrerror(char error_text[]);
inline double *vector(long nl, long nh);
inline void free_vector(double *v, long nl, long nh);
inline void linmin(double p[], double xi[], int n, double *fret, double (*func) (double[]));
inline double brent(double ax, double bx, double cx, double (*f) (double), double tol, double *xmin);
inline double f1dim(double x);
inline void mnbrak(double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double (*func) (double));


/*-----------------------------------------------------------------------------------------------------*
 * Funcoes globais                                                                                     *
 *-----------------------------------------------------------------------------------------------------*/
void AlocarMemoriaRecipes(const int iTamanhoMax)
{
  iTamanhoMaximo = iTamanhoMax;
  pt = vector(1, iTamanhoMaximo);
  ptt = vector(1, iTamanhoMaximo);
  xit = vector(1, iTamanhoMaximo);
  pcom = vector(1, iTamanhoMaximo);
  xicom = vector(1, iTamanhoMaximo);
  xt = vector(1, iTamanhoMaximo);
}


void LiberarMemoriaRecipes()
{
  free_vector(xit, 1, iTamanhoMaximo);
  free_vector(ptt, 1, iTamanhoMaximo);
  free_vector(pt, 1, iTamanhoMaximo);
  free_vector(pcom, 1, iTamanhoMaximo);
  free_vector(xicom, 1, iTamanhoMaximo);
  free_vector(xt, 1, iTamanhoMaximo);
}


void powell(double p[], double **xi, int n, double ftol, int *iter, double *fret, double (*func) (double[]))
{
  int i, ibig, j;
  double del, fp, fptt, t;//, *pt, *ptt, *xit;

  //pt = vector(1, n);
  //ptt = vector(1, n);
  //xit = vector(1, n);
  *fret = (*func) (p);
  for (j = 1; j <= n; j++)
    pt[j] = p[j];
  for (*iter = 1;; ++(*iter)) {
    fp = (*fret);
    ibig = 0;
    del = 0.0;
    for (i = 1; i <= n; i++) {
      for (j = 1; j <= n; j++)
        xit[j] = xi[j][i];
      fptt = (*fret);
      linmin(p, xit, n, fret, func);
      if (fabs(fptt - (*fret)) > del) {
        del = fabs(fptt - (*fret));
        ibig = i;
      }
    }
    if ((2.0*fabs(fp-(*fret)) <= ftol*(fabs(fp)+fabs(*fret))) || (*iter >= ITMAX_POWELL)) {
      //free_vector(xit, 1, n);
      //free_vector(ptt, 1, n);
      //free_vector(pt, 1, n);
      return;
    }
    for (j = 1; j <= n; j++) {
      ptt[j] = 2.0 * p[j] - pt[j];
      xit[j] = p[j] - pt[j];
      pt[j] = p[j];
    }
    fptt = (*func) (ptt);
    if (fptt < fp) {
      t = 2.0 * (fp - 2.0 * (*fret) + fptt) * DSQR(fp - (*fret) - del) - del * DSQR(fp - fptt);
      if (t < 0.0) {
        linmin(p, xit, n, fret, func);
        for (j = 1; j <= n; j++) {
          xi[j][ibig] = xi[j][n];
          xi[j][n] = xit[j];
        }
      }
    }
  }
}


/*-----------------------------------------------------------------------------------------------------*
 * Funcoes locais                                                                                      *
 *-----------------------------------------------------------------------------------------------------*/
inline void nrerror(char error_text[])
{
  fprintf(stderr, "Numerical Recipes run-time error...\n");
  fprintf(stderr, "%s\n", error_text);
  fprintf(stderr, "...now exiting to system...\n");
  //exit(1);
}


inline double *vector(long nl, long nh)
{
  double *v;

  v = (double *) malloc((size_t) ((nh - nl + 1 + NR_END) * sizeof(double)));
  if (!v)
    nrerror("allocation failure in vector()");
  return v - nl + NR_END;
}


inline void free_vector(double *v, long nl, long nh)
{
  free((FREE_ARG) (v + nl - NR_END));
}


inline void linmin(double p[], double xi[], int n, double *fret, double (*func) (double[]))
{
  register int j;
  double xx, xmin, fx, fb, fa, bx, ax;

  ncom = n;
  //pcom = vector(1, n);
  //xicom = vector(1, n);
  nrfunc = func;
  for (j = 1; j <= n; j++) {
    pcom[j] = p[j];
    xicom[j] = xi[j];
  }
  ax = 0.0;
  xx = 1.0;
  mnbrak(&ax, &xx, &bx, &fa, &fx, &fb, f1dim);
  *fret = brent(ax, xx, bx, f1dim, TOL, &xmin);
  for (j = 1; j <= n; j++) {
    xi[j] *= xmin;
    p[j] += xi[j];
  }
  //free_vector(xicom, 1, n);
  //free_vector(pcom, 1, n);
}


inline double brent(double ax, double bx, double cx, double (*f) (double), double tol, double *xmin)
{
  register int iter;
  double a, b, etemp, fu, fv, fw, fx, p, q, r, tol1, tol2, u, v, w, x, xm;
  double d = 0.0, e = 0.0;

  a = (ax < cx ? ax : cx);
  b = (ax > cx ? ax : cx);
  x = w = v = bx;
  fw = fv = fx = (*f) (x);
  for (iter = 1; iter <= ITMAX_BRENT; iter++) {
    xm = 0.5 * (a + b);
    tol2 = 2.0 * (tol1 = tol * fabs(x) + ZEPS);
    if (fabs(x - xm) <= (tol2 - 0.5 * (b - a))) {
      *xmin = x;
      return fx;
    }
    if (fabs(e) > tol1) {
      r = (x - w) * (fx - fv);
      q = (x - v) * (fx - fw);
      p = (x - v) * q - (x - w) * r;
      q = 2.0 * (q - r);
      if (q > 0.0)
        p = -p;
      q = fabs(q);
      etemp = e;
      e = d;
      if (fabs(p) >= fabs(0.5 * q * etemp) || p <= q * (a - x) || p >= q * (b - x))
        d = CGOLD * (e = (x >= xm ? a - x : b - x));
      else {
        d = p / q;
        u = x + d;
        if (u - a < tol2 || b - u < tol2)
          d = SIGN(tol1, xm - x);
      }
    }
    else {
      d = CGOLD * (e = (x >= xm ? a - x : b - x));
    }
    u = (fabs(d) >= tol1 ? x + d : x + SIGN(tol1, d));
    fu = (*f) (u);
    if (fu <= fx) {
      if (u >= x)
        a = x;
      else
        b = x;
      SHFT(v, w, x, u)
          SHFT(fv, fw, fx, fu)
    }
    else {
      if (u < x)
        a = u;
      else
        b = u;
      if (fu <= fw || w == x) {
        v = w;
        w = u;
        fv = fw;
        fw = fu;
      }
      else if (fu <= fv || v == x || v == w) {
        v = u;
        fv = fu;
      }
    }
  }
  *xmin = x;
  return fx;
}


inline double f1dim(double x)
{
  register int j;
  double f;//, *xt;

  //xt = vector(1, ncom);
  for (j = 1; j <= ncom; j++)
    xt[j] = pcom[j] + x * xicom[j];
  f = (*nrfunc) (xt);
  //free_vector(xt, 1, ncom);
  return f;
}


inline void mnbrak(double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double (*func) (double))
{
  double ulim, u, r, q, fu, dum;

  *fa = (*func) (*ax);
  *fb = (*func) (*bx);
  if (*fb > *fa) {
    SHFT(dum, *ax, *bx, dum)
        SHFT(dum, *fb, *fa, dum)
  }
  *cx = (*bx) + GOLD * (*bx - *ax);
  *fc = (*func) (*cx);
  while (*fb > *fc) {
    r = (*bx - *ax) * (*fb - *fc);
    q = (*bx - *cx) * (*fb - *fa);
    u = (*bx) - ((*bx - *cx) * q - (*bx - *ax) * r) / (2.0 * SIGN(DMAX(fabs(q - r), TINY), q - r));
    ulim = (*bx) + GLIMIT * (*cx - *bx);
    if ((*bx - u) * (u - *cx) > 0.0) {
      fu = (*func) (u);
      if (fu < *fc) {
        *ax = (*bx);
        *bx = u;
        *fa = (*fb);
        *fb = fu;
        return;
      }
      else if (fu > *fb) {
        *cx = u;
        *fc = fu;
        return;
      }
      u = (*cx) + GOLD * (*cx - *bx);
      fu = (*func) (u);
    }
    else if ((*cx - u) * (u - ulim) > 0.0) {
      fu = (*func) (u);
      if (fu < *fc) {
        SHFT(*bx, *cx, u, *cx + GOLD * (*cx - *bx))
            SHFT(*fb, *fc, fu, (*func) (u))
      }
    }
    else if ((u - ulim) * (ulim - *cx) >= 0.0) {
      u = ulim;
      fu = (*func) (u);
    }
    else {
      u = (*cx) + GOLD * (*cx - *bx);
      fu = (*func) (u);
    }
    SHFT(*ax, *bx, *cx, u)
        SHFT(*fa, *fb, *fc, fu)
  }
}


#undef NRANSI
